/*****************************************************************************/
/*                                                                           */
/* Description: VQ routines.                                                 */
/* File: "codebook.c"                                                        */
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 9/1/97                                                     */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "matrix.h"
#include "file_io.h"
#include "gen_math.h"
#include "codebook.h"

void LBG_one(LBG_t *local, double epsilon, double delta_dist, int split)
{
  int count = 0;
  double prev_dist, current_dist;

  if(split > 0) 
    L(M) = split_centroid_rand(L(centroid), L(length), L(M), epsilon);

  /*
    display_matrix(stdout, L(centroid), L(M), 2, 6);
    fprintf(stdout, "\n\n");
    */

  prev_dist = total_distortion(local);

  classify(local);
  calculate_centroid(local, epsilon);
  
  /*
    display_matrix(stdout, L(centroid), L(M), 2, 6);
    fprintf(stdout, "\n\n");
    */

  while(fabs((current_dist = total_distortion(local)) -  
	     prev_dist)/prev_dist > delta_dist) {
    fprintf(stderr, "Iteration # %d : %lf\n", ++count, sqrt(current_dist));
    prev_dist = current_dist;

    classify(local);
    calculate_centroid(local, epsilon);

    /*
      display_matrix(stdout, L(centroid), L(M), 2, 6);
      fprintf(stdout, "\n\n");
      */
  }
}

int *int_copy_vector(int *old_vector, int *new_vector, int n)
{
  int i;

  for(i = 0; i < n; i++)
    new_vector[i] = old_vector[i];

  return(new_vector);
}

/* input: data, n, length, M */

void initialize_centroid(LBG_t *local)
{
  L(class) = int_allocate_vector(L(class), L(n));
  L(class) = int_zero_vector(L(class), L(n));
  
  L(centroid) = allocate_matrix(L(centroid), L(M), L(length));
  L(centroid) = zero_matrix(L(centroid), L(M), L(length));

  L(M) = 1;
  calculate_centroid(local, 1.);
}

void calculate_centroid(LBG_t *local, double epsilon)
{
  int i, j, *cnt, flag = 0;

  cnt = int_allocate_vector(cnt, L(M));
  cnt = int_zero_vector(cnt, L(M));
  zero_matrix(L(centroid), L(M), L(length));

  for(i = 0; i < L(n); i++) {
    cnt[L(class[i])]++;
    for(j = 0; j < L(length); j++)
      L(centroid[L(class[i])][j]) += L(data[i][j]);
  }

  for(i = 0; i < L(M); i++)
    if (cnt[i] == 0) {
      flag++;
    } else {
      for(j = 0; j < L(length); j++)
	L(centroid[i][j]) /= (double) cnt[i];
    }

  if(flag > 0) {
    fprintf(stderr, "%d: %d empty centroids\n", L(M), flag);
    reassign_centroid(local, cnt);
  }

  free(cnt);
}

void reassign_centroid(LBG_t *local, int *empty)
{
  int i, j, max_class, vec;
  double *dist, max;

  dist = allocate_vector(dist, L(M));

  for(i = 0; i < L(M); i++) {
    if(empty[i] == 0) {
      dist = class_distortion(local, dist);
      max = -1.;
      for(j = 0; j < L(M); j++)
	if(dist[j] > max) {
	  max_class = j;
	  max = dist[j];
	}
      vec = max_class_distortion(local, max_class);
      copy_vector(L(data[vec]), L(centroid[i]), L(length));
      L(class[vec]) = i;

      for(j = 0; j < L(length); j++)
	L(centroid[max_class][j]) =
	  (L(centroid[max_class][j])*((double) empty[max_class]) -
	   L(data[vec][j]))/((double) (empty[max_class] - 1));
    }
  }

  free(dist);
  return;
}

int max_class_distortion(LBG_t *local, int class)
{
  int max_int, i;
  double max = -1., tmp;
  
  for(i = 0; i < L(n); i++) {
    if(class = L(class[i])) {
      tmp = distortion(L(data[i]), L(centroid[class]), L(length));
      if (tmp > max) {
	max_int = i;
	max = tmp;
      }
    }
  }
  
  return(max_int);
}

double *class_distortion(LBG_t *local, double *dist)
{
  int i, *n;


  n = int_allocate_vector(n, L(M));

  for(i = 0; i < L(M); i++) {
    dist[i] = 0.;
    n[i] = 0;
  }

  for(i = 0; i < L(n); i++) {
    dist[L(class[i])] += 
      distortion(L(data[i]), L(centroid[L(class[i])]), L(length));
    n[L(class[i])]++;
  }
  
  for(i = 0; i < L(M); i++) {
    if(n[i] > 0) {
      dist[i] /= ((double) n[i]);
    }
  }

  free(n);
  return(dist);
}

double total_distortion(LBG_t *local)
{
  int i;
  double dist = 0.;

  for(i = 0; i < L(n); i++)
    dist += distortion(L(data[i]), L(centroid[L(class[i])]), L(length));

  return(dist);
}

double distortion(double *vec1, double *vec2, int length)
{
  int i;
  double sum = 0.;

  for(i = 0; i < length; i++)
      sum += (vec1[i] - vec2[i])*(vec1[i] - vec2[i]);

  return(sum);
}

void classify(LBG_t *local)
{
  int i;

  for(i = 0; i < L(n); i++)
    L(class[i]) = classify_one(L(data[i]), L(length), L(centroid), L(M));
}

int classify_one(double *data_vec, int length, double **centroid, int M)
{
  int i, best_class;
  double current_dist, best_dist;
  
  for(i = 0; i < M; i++) {
    current_dist = distortion(data_vec, centroid[i], length);
    if(i == 0 || current_dist < best_dist) {
      best_dist = current_dist;
      best_class = i;
    }
  }

  return(best_class);
}

int split_centroid(double **centroid, int length, int M, double epsilon)
{
  int i;

  for(i = 0; i < M; i++) {
    add_epsilon_vec(centroid[i+M], centroid[i], length, epsilon);
    add_epsilon_vec(centroid[i], centroid[i], length, -epsilon);
  }

  return(2*M);
}

void add_epsilon_vec(double *vec1, double *vec2, int n, double epsilon)
{
  int i;
  
  for(i = 0; i < n; i++)
    vec1[i] = vec2[i] + epsilon;
}

int split_centroid_rand(double **centroid, int length, int M, double epsilon)
{
  int i;

  double *pp, *pn;

  pp = allocate_vector(pp, length);
  pn = allocate_vector(pn, length);

  for(i = 0; i < length; i++) {
    pp[i] = epsilon * ((double) (random_integer2(0,1)*2-1));
    pn[i] = (-pp[i]);
  }

  for(i = 0; i < M; i++) {
    add_epsilon_vec_rand(centroid[i+M], centroid[i], length, pp);
    add_epsilon_vec_rand(centroid[i], centroid[i], length, pn);
  }

  free(pp);
  free(pn);

  return(2*M);
}

void add_epsilon_vec_rand(double *vec1, double *vec2, int n, double *p)
{
  int i;
  
  for(i = 0; i < n; i++)
    vec1[i] = vec2[i] + p[i];
}

void k_means(LBG_t *local)
{
  int i, j, index, *cnt;

  cnt = int_allocate_vector(cnt, L(M));

  for(i = 0; i < L(M); i++) {
    L(centroid[i]) = copy_vector(L(data[i]), L(centroid[i]), L(length));
    L(class[i]) = i;
    cnt[i] = 1;
  }

  for(i = L(M); i < L(n); i++) {
    index = classify_one(L(data[i]), L(length), L(centroid), L(M));
    L(class[i]) = index;
    for(j = 0; j < L(length); j++)
      L(centroid[index][j]) = (((double) cnt[index])*L(centroid[index][j]) +
			       L(data[i][j]))/((double) cnt[index] + 1.);
    cnt[index]++;
  }

  free(cnt);
}


LBG_t *read_centroid_file(LBG_t *local, char *codebook_file, int vec_length)
{
  FILE *fp;

  L(class) = int_allocate_vector(L(class), L(n));
  L(class) = int_zero_vector(L(class), L(n));
  
  L(M) = count_lines(codebook_file);
  L(length) = vec_length;
  L(centroid) = allocate_matrix(L(centroid), L(M)*2, L(length));

  fp = fopen(codebook_file, "r");
  L(centroid) = read_matrix(fp, L(centroid),L(M), L(length));
  fclose(fp);

  return(local);
}

LBG_t *adjust_vectors_by_q(LBG_t *local, char *q_file)
{
  int i, j;
  FILE *fp;

  if(q_file == NULL)
    return(local);
  else {
    fp = fopen(q_file, "r");
    if(fp == NULL)
      return(local);
    else {
      L(q) = allocate_vector(L(q), L(length));
      read_vector(fp, L(q), L(length));
      fclose(fp);

      for(j = 0; j < L(length); j++)
	L(q[j]) = sqrt(fabs(L(q[j])));

      for(i = 0; i < L(n); i++)
	for(j = 0; j < L(length); j++)
	  L(data[i][j]) *= L(q[j]);
    }
  }

  return(local);
}

#define FILTER_MAX 1024

LBG_t *pick_columns(LBG_t *local, char *in_file, char *filter_file)
{
  FILE *fp;
  int filter[FILTER_MAX], i, j, count = 0, filter_length;
  double tmp;

  if(filter_file == NULL) {
    L(data) = allocate_matrix(L(data), L(n), L(length));
    fp = fopen(in_file, "r");
    read_matrix(fp, L(data), L(n), L(length));
    fclose(fp);
  } else {

    for(i = 0; i < FILTER_MAX; i++)
      filter[i] = 0;

    fp = fopen(filter_file, "r");
    while (fscanf(fp, "%d", &filter[count]) != EOF)
      count++;
    fclose(fp);

    filter_length = count;

    L(length) = 0;
    for(i = 0; i < filter_length; i++) {
      if(filter[i] != 0)
	L(length)++;
    }

    L(data) = allocate_matrix(L(data), L(n), L(length));

    fp = fopen(in_file, "r");

    for(i = 0; i < L(n); i++) {
      count = 0;
      if((i+1) % 1000 == 0)
	fprintf(stderr, 
		"pick_columns(): Done reading %d of %d lines.\n", i+1, L(n));
      for(j = 0; j < filter_length; j++) {
	fscanf(fp, "%lf", &tmp);
	if(filter[j] != 0) {
	  L(data[i][count]) = tmp;
	  count++;
	}
      }
    }
    
    fclose(fp);
  }
  
  return(local);
}
