/*****************************************************************************/
/*                                                                           */
/* Description: Generates VQ codebooks of increasing size.                   */
/* File: "codebook.c"                                                        */
/*                                                                           */
/* Author: Michael C. Nechyba                                                */
/* Last modified: 9/1/97                                                     */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include "matrix.h"
#include "file_io.h"
#include "gen_math.h"
#include "codebook.h"

#define LEVELS_MAX 1024

int main(int argc, char **argv)
{
  FILE *fp;
  LBG_t *local;
  char *in_file = NULL, *out_file = NULL, *q_file = NULL, file[STR_REG];
  char *filter_file = NULL, *codebook_file = NULL;
  int i, j, M_max= LEVELS_MAX, n_max = 7, length = 16, precision = 15;
  int seed = 1;
  double epsilon = .0001, delta_dist = .01, dis, mag, last = 10.;

  arg_parse(argc, argv,
	    "", "Usage:",
	    "-input %S", &in_file,  "set input file",
	    "-output %S", &out_file, "set output file",
	    "-initial [%S]", &codebook_file,
	    "set initialization codebook file to read [none]",
	    "-convergence [%F]", &delta_dist, 
	    "set convergence criterion [%lf]", delta_dist,
	    "-last [%F]", &last,
	    "divide convergence criterion by [%lf] on last split", last,
	    "-qFile [%S]", &q_file, "set Q file [none]",
	    "-epsilon [%F]", &epsilon, "set epsilon [%lf]", epsilon,
	    "-filter [%S]", &filter_file, "set filter file [none]",
	    "-length [%d]", &length, "length of each vector [%d]", length,
	    "-nCodes [%d]", &n_max, "set number of codebooks [2^%d]", n_max,
	    "-precision [%d]", &precision, 
	    "set output precision [%d]", precision,
	    "-seed [%d]", &seed, "random seed [%d]", seed,
	    NULL);

  srand(seed);

  if(in_file == NULL || out_file == NULL) {
    fprintf(stderr, "No input and/or output file specified.\n");
    exit(-1);
  }

  local = (LBG_t *) malloc(sizeof(LBG_t));
  is_null((int *) local);

  L(n) = count_lines(in_file);
  L(length) = length;

  pick_columns(local, in_file, filter_file);
  adjust_vectors_by_q(local, q_file);

  L(M) = M_max;

  if(codebook_file == NULL) {
    initialize_centroid(local);
    
    dis = total_distortion(local);
    fprintf(stdout, 
	    "Distortion for 1 code: (Total = %lf ), (normalized = %.10f )\n", 
	    sqrt(dis), sqrt(dis/((double) (L(length)*L(n)))));
    
    for(i = 0; i < n_max; i++) {
      if(i == n_max - 1)
	delta_dist /= last;
      LBG_one(local, epsilon, delta_dist, 1);
      dis = total_distortion(local);
      fprintf(stdout, 
	      "Distortion for %d codes: Total = %lf , normalized = %.10f\n", 
	      (int) pow(2., (double) (i+1)), 
	      sqrt(dis), sqrt(dis/((double) (L(length)*L(n)))));
      
      sprintf(file, "%s.%d", out_file, i+1);
      
      fp = fopen(file, "w");
      for(j = 0; j < L(M); j++) {
	mag = abs_vec(L(centroid[j]), L(length));
	if(mag > .0000000001)
	  display_vector(fp, L(centroid[j]), L(length), 0, precision);
      }
      fclose(fp);
    }
  } else {
    read_centroid_file(local, codebook_file, L(length));
    LBG_one(local, epsilon, delta_dist, -1);

    dis = total_distortion(local);

    fprintf(stdout, 
	    "Distortion for %d codes: Total = %lf , normalized = %.10f\n", 
	    L(M), sqrt(dis), sqrt(dis/((double) (L(length)*L(n)))));
      
    fp = fopen(out_file, "w");
    for(j = 0; j < L(M); j++) {
      mag = abs_vec(L(centroid[j]), L(length));
      if(mag > .0000000001)
	display_vector(fp, L(centroid[j]), L(length), 0, precision);
    }
    fclose(fp);
    
    /*
    LBG_one(local, epsilon, delta_dist, 1);

    dis = total_distortion(local);

    fprintf(stdout, 
	    "Distortion for %d codes: Total = %lf , normalized = %.10f\n", 
	    L(M), sqrt(dis), sqrt(dis/((double) (L(length)*L(n)))));
      
    fp = fopen(out_file, "w");
    for(j = 0; j < L(M); j++) {
      mag = abs_vec(L(centroid[j]), L(length));
      if(mag > .0000000001)
	display_vector(fp, L(centroid[j]), L(length), 0, precision);
    }
    fclose(fp);
    */
  }
}
